cmake_minimum_required(VERSION 2.8)

# Test the PYEDEV_CONFIG command line argument on inclusion.
# Attempt to include the file if string not empty.
if("${PYEDEV_CONFIG}" STREQUAL "")
  message(FATAL_ERROR "PYEDEV_CONFIG is not defined")
endif()
include("${PYEDEV_CONFIG}")

# Set default build platform type variables.
set(PYEDEV_BUILD_PLATFORM_CROSS NO)
set(PYEDEV_BUILD_PLATFORM_NATIVE NO)
set(PYEDEV_BUILD_PLATFORM_NONE NO)

# Test the PYEDEV_PLATFORM command line argument on inclusion.
# CMake does not seem able to use multiple toolchains during a build, so this
# switch differentiates between a cross compiler toolchain (e.g. GNU ARM), or
# a native one. Library and binary targets are built when cross-compiling, and
# test targets are ignored, compiling for the native system results in reverse
# behaviour (libraries/binaries ignored, tests built). There is also a none
# option for compiler independent targets such as documentation generators.
if("${PYEDEV_PLATFORM}" STREQUAL "")
  message(FATAL_ERROR "PYEDEV_PLATFORM is not defined (cross|native)")
elseif("${PYEDEV_PLATFORM}" STREQUAL "cross")
  # Cross platform selected, test targets ignored.
  set(PYEDEV_BUILD_PLATFORM_CROSS YES)
elseif("${PYEDEV_PLATFORM}" STREQUAL "native")
  # Native platform selected, library and binary targets ignored.
  set(PYEDEV_BUILD_PLATFORM_NATIVE YES)
elseif("${PYEDEV_PLATFORM}" STREQUAL "none")
  # None platform selected, library, binary and test targets ignored.
  set(PYEDEV_BUILD_PLATFORM_NONE YES)
else()
  message(FATAL_ERROR "PYEDEV_PLATFORM is invalid (cross|native)")
endif()

# Test the PYEDEV_BUILD_TYPE command line argument on inclusion.
# This switch differentiates between debug/release builds, and is used by
# install directives to separate libraries/binaries that have been built
# using different levels of code optimisation.
if("${PYEDEV_BUILD_TYPE}" STREQUAL "")
  message(FATAL_ERROR "PYEDEV_BUILD_TYPE is not defined (debug|release)")
elseif("${PYEDEV_BUILD_TYPE}" STREQUAL "debug")
  # Do nothing if valid.
elseif("${PYEDEV_BUILD_TYPE}" STREQUAL "release")
  # Do nothing if valid.
else()
  message(FATAL_ERROR "PYEDEV_BUILD_TYPE is invalid (debug|release)")
endif()

# Functions/macros...

# TODO: Debug/release build/install directories.

function(pyedev_build_util_configure_files ARG_FILES)
  if("${ARG_FILES}" STREQUAL "")
    message(FATAL_ERROR "File list empty")
  else()
    # START pyedev_build_util_configure_files

    # Iterate over input files, configuring them if they exist.
    foreach(INPUT_FILE ${ARG_FILES})
      if(EXISTS "${INPUT_FILE}.in")
        configure_file("${INPUT_FILE}.in" ${INPUT_FILE} @ONLY)
      endif()
    endforeach(INPUT_FILE)

    # END pyedev_build_util_configure_files
  endif()
endfunction(pyedev_build_util_configure_files)

macro(pyedev_build_util_arch_flags ARG_ARCH OUT_COMPILER_FLAGS OUT_LD_FLAGS)
  if("${ARG_ARCH}" STREQUAL "")
    message(FATAL_ERROR "No target architecture")
  elseif("${OUT_COMPILER_FLAGS}" STREQUAL "")
    message(FATAL_ERROR "No output compiler flags variable")
  elseif("${OUT_LD_FLAGS}" STREQUAL "")
    message(FATAL_ERROR "No output linker flags variable")
  else()
    # START pyedev_build_util_arch_flags

    # Set output flag variables based on architecture argument.
    if("${ARG_ARCH}" STREQUAL "cortex-m7")
      # ARM Cortex-M7 compiler and linker flags.
      set(${OUT_COMPILER_FLAGS} "-mthumb -mcpu=cortex-m7 -mfloat-abi=softfp -mfpu=fpv4-sp-d16")
      set(${OUT_LD_FLAGS} "-mthumb -mcpu=cortex-m7")
    elseif("${ARG_ARCH}" STREQUAL "cortex-m4")
      # ARM Cortex-M4 compiler and linker flags.
      set(${OUT_COMPILER_FLAGS} "-mthumb -mcpu=cortex-m4 -mfloat-abi=softfp -mfpu=fpv4-sp-d16")
      set(${OUT_LD_FLAGS} "-mthumb -mcpu=cortex-m4")
    elseif("${ARG_ARCH}" STREQUAL "cortex-m0")
      # ARM Cortex-M0 compiler and linker flags.
      set(${OUT_COMPILER_FLAGS} "-mthumb -mcpu=cortex-m0")
      set(${OUT_LD_FLAGS} "-mthumb -mcpu=cortex-m0")
    else()
      # Error on invalid architecture argument.
      message(FATAL_ERROR "Invalid target architecture")
    endif()

    # END pyedev_build_util_arch_flags
  endif()
endmacro(pyedev_build_util_arch_flags)

# Create a static library headers build target.
# Arguments:
#  ARG_PREFIX:  Unique string to prefix library headers target with.
#  ARG_HEADERS: Target library header files list.
#  ARG_INSTALL: Installation location for library headers.
#  OUT_TARGET:  Output variable name set to custom target.
macro(pyedev_build_target_library_headers ARG_PREFIX ARG_HEADERS ARG_INSTALL OUT_TARGET)
    if("${ARG_PREFIX}" STREQUAL "")
        message(FATAL_ERROR "No target prefix argument")
    elseif("${ARG_HEADERS}" STREQUAL "")
        message(FATAL_ERROR "Headers file list empty")
    elseif("${ARG_INSTALL}" STREQUAL "")
        message(FATAL_ERROR "Install directory undefined")
    elseif("${OUT_TARGET}" STREQUAL "")
        message(FATAL_ERROR "Output target name undefined")
    else()
        # Create a custom target for library headers which creates install directory.
        add_custom_target("${ARG_PREFIX}-headers" ALL
            COMMAND ${CMAKE_COMMAND} -E make_directory "${ARG_INSTALL}")

        # For each header file, create custom command which copies it to installation
        # directory without having to invoke install. So if necessary it can be found
        # by targets dependant on these library headers.
        foreach(FILE_I ${ARG_HEADERS})
            add_custom_command(TARGET "${ARG_PREFIX}-headers" POST_BUILD
                COMMAND ${CMAKE_COMMAND} -E copy "${FILE_I}" "${ARG_INSTALL}")

            # Add installation target for header file.
            install(FILES ${FILE_I} DESTINATION "${ARG_INSTALL}")
        endforeach(FILE_I)

        # Set the output target variable to created target.
        set("${OUT_TARGET}" "${ARG_PREFIX}-headers")
    endif()
endmacro(pyedev_build_target_library_headers)

# Create a static library archive build target.
# Arguments:
#  ARG_TARGET:         Target library name string.
#  ARG_ARCH:           Target architecture string.
#  ARG_COMPILER_FLAGS: Optional target compiler flags.
#  ARG_INC:            Target include directories list.
#  ARG_SRC:            Target source files list.
#  ARG_DEPENDS:        List of target dependencies.
#  ARG_INSTALL:        Installation location for archive output.
#  OUT_TARGET:         Variable name set to created library target.
macro(pyedev_build_target_library ARG_TARGET ARG_ARCH ARG_COMPILER_FLAGS ARG_INC ARG_SRC ARG_DEPENDS
                                  ARG_INSTALL OUT_TARGET)
    # Check the platform type is CROSS, else ignore target.
    if(${PYEDEV_BUILD_PLATFORM_CROSS})
        if("${ARG_TARGET}" STREQUAL "")
            message(FATAL_ERROR "Empty target name argument")
        # ARG_ARCH argument checked by pyedev_build_util_arch_flags
        # ARG_COMPILER_FLAGS is optional.
        # ARG_INC is optional.
        elseif("${ARG_SRC}" STREQUAL "")
            message(FATAL_ERROR "Source files list empty")
        # ARG_DEPENDS is optional.
        elseif("${ARG_INSTALL}" STREQUAL "")
            message(FATAL_ERROR "Empty install directory argument")
        elseif("${OUT_TARGET}" STREQUAL "")
            message(FATAL_ERROR "Output target name undefined")
        else()
            # Get compiler flags for the target architecture.
            pyedev_build_util_arch_flags("${ARG_ARCH}" "_PYEDEV_ARCH_COMPILER_FLAGS" "_PYEDEV_ARCH_LD_FLAGS")

            # Combine architecture and compiler flags argument.
            set("_PYEDEV_COMPILER_FLAGS" "${_PYEDEV_ARCH_COMPILER_FLAGS} ${ARG_COMPILER_FLAGS}")

            # Set CMake compiler flags using configuration file variables.
            set(CMAKE_C_FLAGS "${CMAKE_C_FLAGS} ${PYEDEV_CONFIG_C_FLAGS} ${PYEDEV_CONFIG_COMPILER_FLAGS}")
            set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} ${PYEDEV_CONFIG_CXX_FLAGS} ${PYEDEV_CONFIG_COMPILER_FLAGS}")

            # Add static library target using name and source list arguments.
            add_library("${ARG_TARGET}" STATIC ${ARG_SRC})

            # Set library target compiler flags.
            set_target_properties("${ARG_TARGET}" PROPERTIES COMPILE_FLAGS "${_PYEDEV_COMPILER_FLAGS}")

            # Set library target include directories (include configuration install includes by default).
            target_include_directories("${ARG_TARGET}" PUBLIC "${PYEDEV_CONFIG_INSTALL_INC};${ARG_INC}")

            # Add library target dependencies.
            foreach(DEPEND_I ${ARG_DEPENDS})
                add_dependencies("${ARG_TARGET}" "${DEPEND_I}")
            endforeach(DEPEND_I)

            # For created library archive, create custom command which copies it to installation
            # directory without having to invoke install. So if necessary it can be found by
            # targets dependant on this library.
            add_custom_command(TARGET "${ARG_TARGET}" POST_BUILD
                COMMAND ${CMAKE_COMMAND} -E make_directory "${ARG_INSTALL}"
                COMMAND ${CMAKE_COMMAND} -E copy "${CMAKE_CURRENT_BINARY_DIR}/lib${ARG_TARGET}.a" "${ARG_INSTALL}")

            # Add library archive installation target.
            install(FILES "${CMAKE_CURRENT_BINARY_DIR}/lib${ARG_TARGET}.a" DESTINATION "${ARG_INSTALL}")

            # Set the output target variable to created target.
            set("${OUT_TARGET}" "${ARG_TARGET}")
        endif()
    endif()
endmacro(pyedev_build_target_library)

macro(pyedev_build_target_binary ARG_TARGET ARG_ARCH ARG_COMPILER_FLAGS
                                 ARG_LD_SCRIPT ARG_LD_FLAGS ARG_INC
                                 ARG_SRC ARG_LIBS ARG_DEPENDS ARG_INSTALL
                                 OUT_TARGET)
  # Check the platform is cross, else ignore target.
  if(${PYEDEV_BUILD_PLATFORM_CROSS})
    if("${ARG_TARGET}" STREQUAL "")
      message(FATAL_ERROR "No target name")
    # ARG_ARCH argument checked by pyedev_build_util_arch_flags
    # ARG_COMPILER_FLAGS is optional.
    elseif("${ARG_LD_SCRIPT}" STREQUAL "")
      message(FATAL_ERROR "No linker script")
    # ARG_LD_FLAGS is optional.
    # ARG_INC is optional.
    elseif("${ARG_SRC}" STREQUAL "")
      message(FATAL_ERROR "No source files list")
    # ARG_LIBS is optional.
    # ARG_DEPENDS is optional.
    elseif("${ARG_INSTALL}" STREQUAL "")
      message(FATAL_ERROR "No install directory")
    elseif("${OUT_TARGET}" STREQUAL "")
      message(FATAL_ERROR "No output target variable")
    else()
      # START pyedev_build_target_binary

      # Get compiler flags for the target architecture.
      pyedev_build_util_arch_flags(${ARG_ARCH} _PYEDEV_ARCH_COMPILER_FLAGS _PYEDEV_ARCH_LD_FLAGS)

      # Combine architecture and compiler/linker flags arguments (include
      # configuration file linker flags by default).
      set(_PYEDEV_COMPILER_FLAGS "${_PYEDEV_ARCH_COMPILER_FLAGS} ${ARG_COMPILER_FLAGS}")
      set(_PYEDEV_LD_FLAGS "${_PYEDEV_ARCH_LD_FLAGS} ${PYEDEV_CONFIG_LD_FLAGS} ${ARG_LD_FLAGS}")

      # Set binary output file names.
      set(_PYEDEV_ELF_TARGET "${ARG_TARGET}.${PYEDEV_BUILD_TYPE}.elf")
      set(_PYEDEV_BIN_TARGET "${ARG_TARGET}.${PYEDEV_BUILD_TYPE}.bin")
      set(_PYEDEV_MAP_TARGET "${ARG_TARGET}.${PYEDEV_BUILD_TYPE}.map")

      # Set CMake compiler flags using configuration file variables.
      set(CMAKE_ASM_FLAGS "${CMAKE_ASM_FLAGS} ${PYEDEV_CONFIG_ASM_FLAGS} ${PYEDEV_CONFIG_COMPILER_FLAGS}")
      set(CMAKE_C_FLAGS "${CMAKE_C_FLAGS} ${PYEDEV_CONFIG_C_FLAGS} ${PYEDEV_CONFIG_COMPILER_FLAGS}")
      set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} ${PYEDEV_CONFIG_CXX_FLAGS} ${PYEDEV_CONFIG_COMPILER_FLAGS}")

      # Add binary executable target using input target and sources.
      add_executable(${_PYEDEV_ELF_TARGET} ${ARG_SRC})

      # Set binary target compiler and linker flags (also set linker
      # script argument and map output).
      set_target_properties(${_PYEDEV_ELF_TARGET} PROPERTIES COMPILE_FLAGS ${_PYEDEV_COMPILER_FLAGS})
      set_target_properties(${_PYEDEV_ELF_TARGET} PROPERTIES LINK_FLAGS "-T${ARG_LD_SCRIPT} -Wl,-M=${_PYEDEV_MAP_TARGET} ${_PYEDEV_LD_FLAGS}")

      # Set binary target include directories (include configuration
      # install includes by default).
      target_include_directories(${_PYEDEV_ELF_TARGET} PUBLIC "${PYEDEV_CONFIG_INSTALL_INC};${ARG_INC}")

      # Add binary target linked libraries.
      foreach(LIB_I ${ARG_LIBS})
        target_link_libraries(${_PYEDEV_ELF_TARGET} ${LIB_I})
      endforeach(LIB_I)

      # Add binary target dependencies.
      foreach(DEPEND_I ${ARG_DEPENDS})
        add_dependencies(${_PYEDEV_ELF_TARGET} ${DEPEND_I})
      endforeach(DEPEND_I)

      # Absolute output and install file paths.
      set(_PYEDEV_ELF_OUTPUT "${CMAKE_CURRENT_BINARY_DIR}/${_PYEDEV_ELF_TARGET}")
      set(_PYEDEV_BIN_OUTPUT "${CMAKE_CURRENT_BINARY_DIR}/${_PYEDEV_BIN_TARGET}")
      set(_PYEDEV_MAP_OUTPUT "${CMAKE_CURRENT_BINARY_DIR}/${_PYEDEV_MAP_TARGET}")
      set(_PYEDEV_BINARY_INSTALL "${PYEDEV_CONFIG_INSTALL_BIN}/${ARG_INSTALL}")

      # For created binary, create custom command for size and binary output, uses toolchain binaries
      # defined by configuration file. Also copies output files to installation directory without
      # having to invoke install. So if necessary they can be found by dependent targets.
      add_custom_command(TARGET ${_PYEDEV_ELF_TARGET} POST_BUILD
        COMMAND ${PYEDEV_CONFIG_TOOLCHAIN_SIZE} ${CMAKE_CURRENT_BINARY_DIR}/${_PYEDEV_ELF_TARGET}
        COMMAND ${PYEDEV_CONFIG_TOOLCHAIN_OBJCOPY} -O binary ${_PYEDEV_ELF_OUTPUT} ${_PYEDEV_BIN_TARGET}.tmp
        COMMAND ${CMAKE_COMMAND} -E copy ${_PYEDEV_BIN_TARGET}.tmp ${_PYEDEV_BIN_OUTPUT}
        COMMAND ${CMAKE_COMMAND} -E make_directory ${_PYEDEV_BINARY_INSTALL}
        COMMAND ${CMAKE_COMMAND} -E copy ${_PYEDEV_ELF_OUTPUT} ${_PYEDEV_BINARY_INSTALL}
        COMMAND ${CMAKE_COMMAND} -E copy ${_PYEDEV_BIN_OUTPUT} ${_PYEDEV_BINARY_INSTALL}
        COMMAND ${CMAKE_COMMAND} -E copy ${_PYEDEV_MAP_OUTPUT} ${_PYEDEV_BINARY_INSTALL}
        DEPENDS ${_PYEDEV_ELF_TARGET})

      # Redundant install targets, kept so that make install does something.
      # Serves as test to check copy commands are functioning.
      install(FILES ${_PYEDEV_ELF_OUTPUT} DESTINATION ${_PYEDEV_BINARY_INSTALL})
      install(FILES ${_PYEDEV_BIN_OUTPUT} DESTINATION ${_PYEDEV_BINARY_INSTALL})
      install(FILES ${_PYEDEV_MAP_OUTPUT} DESTINATION ${_PYEDEV_BINARY_INSTALL})

      # Set the output target variable to created target.
      set(${OUT_TARGET} ${_PYEDEV_ELF_TARGET})

      # START pyedev_build_target_binary
    endif()
  endif()
endmacro(pyedev_build_target_binary)

macro(pyedev_build_target_test ARG_TARGET ARG_COMPILER_FLAGS ARG_INC ARG_SRC
                               ARG_DEPENDS ARG_INSTALL OUT_TARGET)
  # Check the platform type is native, else ignore target.
  if(${PYEDEV_BUILD_PLATFORM_NATIVE})
    if("${ARG_TARGET}" STREQUAL "")
      message(FATAL_ERROR "No target name")
    # ARG_COMPILER_FLAGS is optional.
    # ARG_INC is optional.
    elseif("${ARG_SRC}" STREQUAL "")
      message(FATAL_ERROR "No source files list")
    # ARG_DEPENDS is optional.
    elseif("${ARG_INSTALL}" STREQUAL "")
      message(FATAL_ERROR "No install directory")
    elseif("${OUT_TARGET}" STREQUAL "")
      message(FATAL_ERROR "No output target variable")
    else()
      # START pyedev_build_target_test

      # Set CMake compiler flags using configuration file variables.
      set(CMAKE_C_FLAGS "${CMAKE_C_FLAGS} ${PYEDEV_CONFIG_C_FLAGS} ${PYEDEV_CONFIG_COMPILER_FLAGS}")
      set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} ${PYEDEV_CONFIG_CXX_FLAGS} ${PYEDEV_CONFIG_COMPILER_FLAGS}")

      # Set test output file name.
      set(_PYEDEV_TEST_TARGET "${ARG_TARGET}.${PYEDEV_BUILD_TYPE}")

      # Create executable output target.
      add_executable(${_PYEDEV_TEST_TARGET} ${ARG_SRC})

      # Set executable target compiler flags.
      set_target_properties(${_PYEDEV_TEST_TARGET} PROPERTIES COMPILE_FLAGS "${ARG_COMPILER_FLAGS}")

      # Set executable target include directories (include configuration install includes by default).
      target_include_directories(${_PYEDEV_TEST_TARGET} PUBLIC "${PYEDEV_CONFIG_INSTALL_INC};${ARG_INC}")

      # Add test target dependencies.
      foreach(DEPEND_I ${ARG_DEPENDS})
        add_dependencies(${_PYEDEV_TEST_TARGET} ${DEPEND_I})
      endforeach(DEPEND_I)

      # Get the output executable target file path (assumes unix or windows).
      if(${WIN32})
        set(_PYEDEV_TEST_OUTPUT "${CMAKE_CURRENT_BINARY_DIR}/${_PYEDEV_TEST_TARGET}.exe")
      else()
        set(_PYEDEV_TEST_OUTPUT "${CMAKE_CURRENT_BINARY_DIR}/${_PYEDEV_TEST_TARGET}")
      endif()
      set(_PYEDEV_TEST_INSTALL "${PYEDEV_CONFIG_INSTALL_TESTS}/${ARG_INSTALL}")

      # For created test execitable, create custom command which copies it to installation
      # directory without having to invoke install. So if necessary it can be found by
      # targets dependant on this test.
      add_custom_command(TARGET ${_PYEDEV_TEST_TARGET} POST_BUILD
        COMMAND ${CMAKE_COMMAND} -E make_directory ${_PYEDEV_TEST_INSTALL}
        COMMAND ${CMAKE_COMMAND} -E copy ${_PYEDEV_TEST_OUTPUT} ${_PYEDEV_TEST_INSTALL}
        COMMAND ${_PYEDEV_TEST_OUTPUT})

      # Redundant install target, kept so that make install does something.
      # Serves as test to check copy commands are functioning.
      install(FILES ${_PYEDEV_TEST_OUTPUT} DESTINATION ${_PYEDEV_TEST_INSTALL})

      # Set the output target variable to created target.
      set(${OUT_TARGET} ${_PYEDEV_TEST_TARGET})

      # END pyedev_build_target_test
    endif()
  endif()
endmacro(pyedev_build_target_test)

macro(pyedev_build_target_doxygen ARG_TARGET ARG_DOXYFILE ARG_OUTPUT ARG_INSTALL)
  # Check the platform is none, else ignore target.
  if(${PYEDEV_BUILD_PLATFORM_NONE})
    if("${ARG_TARGET}" STREQUAL "")
      message(FATAL_ERROR "No target name")
    elseif("${ARG_DOXYFILE}" STREQUAL "")
      message(FATAL_ERROR "No doxygen configuration file")
    elseif("${ARG_OUTPUT}" STREQUAL "")
      message(FATAL_ERROR "No doxygen output directory")
    elseif("${ARG_INSTALL}" STREQUAL "")
      message(FATAL_ERROR "No install directory")
    else()
      # START pyedev_build_target_doxygen

      # Set a variable for testing if Doxygen targets have not been ignored.
      set(DOXYGEN_USED NO)

      # Check Doxygen is installed.
      find_program(DOXYGEN_EXEC doxygen)
      if("${DOXYGEN_EXEC}" STREQUAL "DOXYGEN_EXEC-NOTFOUND")
        message(WARNING "Doxygen is not installed, targets will be ignored")
      else()
        set(DOXYGEN_USED YES)

        # Add Doxygen target with doxyfile input.
        add_custom_target(${ARG_TARGET} ALL
          COMMAND ${DOXYGEN_EXEC} ${ARG_DOXYFILE}
          COMMENT "Doxygen: generating documentation")

        # Absolute output and install directory paths
        # (trailing slash is important!).
        set(_PYEDEV_DOXYGEN_OUTPUT "${CMAKE_CURRENT_BINARY_DIR}/${ARG_OUTPUT}/html/")
        set(_PYEDEV_DOXYGEN_INSTALL "${PYEDEV_CONFIG_INSTALL_DOCS}/${ARG_INSTALL}/html")

        # Add a post build command to copy generated html to install
        # directory. Also copies output html to installation directory
        # without having to invoke install. So if necessary it can be
        # found by dependent targets.
        add_custom_command(TARGET ${ARG_TARGET} POST_BUILD
          COMMAND ${CMAKE_COMMAND} -E copy_directory ${_PYEDEV_DOXYGEN_OUTPUT} ${_PYEDEV_DOXYGEN_INSTALL})

        # Redundant install target, kept so that make install does something.
        # Serves as test to check copy commands are functioning.
        install(DIRECTORY ${_PYEDEV_DOXYGEN_OUTPUT} DESTINATION ${_PYEDEV_DOXYGEN_INSTALL})
      endif()

      # END pyedev_build_target_doxygen
    endif()
  endif()
endmacro(pyedev_build_target_doxygen)
