from os import makedirs
from os.path import abspath, join, isdir
from shutil import rmtree
# Use check_call rather than run for Python 2.7 support.
from subprocess import check_call, CalledProcessError
from pyedev.cmake import BUILD_FILE, TOOLCHAIN_FILE
from pyedev.commands import akw
from pyedev.commands.config import config_list, config_path

BUILD_HELP = "CMake build (re-)generation utility."


def build_cmake_path(*args):
    """Ensure paths are CMake friendly.

    Args:
        path (str): Path string to modify.
    """
    path = akw(0, args, str, None, invalid=("",), required=True)
    return str(path.replace("\\", "/"))


def build_cmake_proc(*args):
    """Run CMake build subprocess for target directory.

    Kwargs:
        target_dir (str): CMake target directory.
        build_dir (str): CMake build directory.
        generator (str): CMake generator type.
        defines (dict): CMake define parameters.
    """
    target_dir = akw(0, args, str, None, invalid=("",), required=True)
    build_dir = akw(1, args, str, None, invalid=("",), required=True)
    generator = akw(2, args, str, None, invalid=("",), required=True)
    defines = akw(3, args, dict, {})

    # If build directory does not exist, attempt to create it.
    if not isdir(build_dir):
        try:
            makedirs(build_dir)
        except OSError as e:
            return bool(False)

    cmake_args = []
    # Append dictionary definitions as defines.
    for define in defines:
        cmake_args.append("-D{0}={1}".format(define, defines[define]))

    # Add generator and target arguments.
    cmake_args.append("-G")
    cmake_args.append(generator)
    cmake_args.append(target_dir)

    # Run CMake subprocess inside build directory.
    try:
        check_call(["cmake"] + cmake_args, cwd=build_dir)
    except CalledProcessError as e:
        return bool(False)

    return bool(True)


def build_make_proc(*args):
    """Run Make subprocess for target directory.

    Args:
        build_dir (str): Working directory for make process.
        make_args (list): Make arguments list.
    """
    build_dir = akw(0, args, str, None, invalid=("",), required=True)
    make_args = akw(1, args, list, [])

    # Run make subprocess inside target directory.
    try:
        check_call(["make"] + make_args, cwd=build_dir)
    except CalledProcessError as e:
        return bool(False)

    return bool(True)


def build_subparser(sb):
    """Populate subparser for build command.

    Arguments:
        sb (obj): argparse.ArgumentParser object.
    """
    # Required configuration name option.
    sb.add_argument("-c", "--config",
                    nargs=1,
                    type=str,
                    choices=config_list(),
                    required=True,
                    help="use configuration",
                    metavar="NAME")

    # Platform type option.
    sb.add_argument("-p", "--platform",
                    nargs=1,
                    default=["all"],
                    type=str,
                    choices=["all", "cross", "native", "none"],
                    help="platform target",
                    metavar="TYPE")

    # Build type option.
    sb.add_argument("-b", "--build",
                    nargs=1,
                    default=["all"],
                    type=str,
                    choices=["all", "debug", "release"],
                    help="build type",
                    metavar="TYPE")

    # Generator option passed directly to CMake.
    sb.add_argument("-G", "--generator",
                    nargs=1,
                    default=["Unix Makefiles"],
                    type=str,
                    help="CMake generator",
                    metavar="NAME")

    # Rebuild flag.
    sb.add_argument("-r", "--rebuild",
                    action="store_true",
                    help="rebuild target")

    # Target directory positional argument.
    sb.add_argument("target",
                    nargs=1,
                    type=str,
                    help="target directory",
                    metavar="DIR")

    # Optional generated makefile arguments.
    sb.add_argument("make",
                    nargs="*",
                    type=str,
                    help="makefile arguments",
                    metavar="MAKE_ARGS")

    # Set callback function.
    sb.set_defaults(func=_subparser_func)


def _subparser_func(args):
    defines = dict({})

    # Absolute CMake configuration file path define.
    config_file = config_path(args.config[0])
    defines["PYEDEV_CONFIG"] = build_cmake_path(config_file)

    # Absolute CMake build file path define.
    build_file = build_cmake_path(BUILD_FILE)
    defines["PYEDEV_BUILD"] = build_file

    # Absolute target directory path define.
    target_dir = build_cmake_path(abspath(args.target[0]))
    defines["PYEDEV_CWD"] = target_dir

    # Determine the platform and build types, if not specified at the
    # command line will default to all.
    platforms = []
    if args.platform[0] == "all":
        platforms += ["cross", "native", "none"]
    elif args.platform[0] == "cross":
        platforms += ["cross"]
    elif args.platform[0] == "native":
        platforms += ["native"]
    else:
        platforms += ["none"]

    builds = []
    if args.build[0] == "all":
        builds += ["debug", "release"]
    elif args.build[0] == "debug":
        builds += ["debug"]
    else:
        builds += ["release"]

    # Build root directory.
    build_root_dir = join(target_dir, ".pyedev")

    # If rebuild flag set, remove existing build directories.
    if args.rebuild:
        if isdir(build_root_dir):
            try:
                rmtree(build_root_dir)
            except OSError as e:
                return tuple((1, str(e)))

    # For each target platform and build type, (re-)generate build
    # using CMake and if Make arguments supplied call generated
    # Makefile with arguments.
    for platform in platforms:
        # The none platform type doesn't depend on build type.
        if platform == "none":
            # Platform type and build directory defines.
            defines["PYEDEV_PLATFORM"] = "none"
            build_dir = join(build_root_dir, "cross.none")
            # Additional toolchain file define to avoid native compiler discovery.
            toolchain_file = build_cmake_path(TOOLCHAIN_FILE)
            defines["CMAKE_TOOLCHAIN_FILE"] = toolchain_file

            # Call CMake build subprocess, return if error occurred.
            if not build_cmake_proc(target_dir, build_dir, args.generator[0], defines):
                emsg = "cmake failed ({0})".format(platform)
                return tuple((1, emsg))

            # If make arguments provided, call Make subprocess.
            if len(args.make) > 0:
                if not build_make_proc(build_dir, args.make):
                    emsg = "make failed ({0})".format(platform)
                    return tuple((1, emsg))

        # The cross and native platform types depend on build type.
        else:
            for build in builds:
                # Platform type.
                defines["PYEDEV_PLATFORM"] = platform
                if platform == "cross":
                    # Additional toolchain file define for cross compilation.
                    toolchain_file = build_cmake_path(TOOLCHAIN_FILE)
                    defines["CMAKE_TOOLCHAIN_FILE"] = toolchain_file
                else:
                    # Remove toolchain file define for native compilation.
                    if "CMAKE_TOOLCHAIN_FILE" in defines:
                        defines.pop("CMAKE_TOOLCHAIN_FILE")

                # Internal and CMake build types.
                defines["PYEDEV_BUILD_TYPE"] = build
                if build == "debug":
                    defines["CMAKE_BUILD_TYPE"] = "Debug"
                else:
                    defines["CMAKE_BUILD_TYPE"] = "MinSizeRel"

                # Build directory based on platform/build types.
                build_dir = join(build_root_dir, "{0}.{1}".format(platform, build))

                # Call CMake build subprocess, return if error occurred.
                if not build_cmake_proc(target_dir, build_dir, args.generator[0], defines):
                    emsg = "cmake failed ({0}.{1})".format(platform, build)
                    return tuple((1, emsg))

                # If make arguments provided, call Make subprocess.
                if len(args.make) > 0:
                    if not build_make_proc(build_dir, args.make):
                        emsg = "make failed ({0}.{1})".format(platform, build)
                        return tuple((1, emsg))

    # Successful build.
    return tuple((0, None))
