from os import listdir, remove
from os.path import splitext, isfile, join
from shutil import copyfile
from pyedev.commands import akw
from pyedev.cmake import CONFIG_DIR, CONFIG_TEMPLATE_FILE

CONFIG_HELP = "CMake configuration file(s) manager."


def config_list():
    """Get list of existing configuration names."""
    d = CONFIG_DIR
    # List of configurations exluding template in cmake/config.
    l = [splitext(f)[0] for f in listdir(d) if isfile(join(d, f))]
    l.remove("template")
    return list(l)


def config_path(*args):
    """Get absolute configuration path from name.

    Args:
        name (str): Configuration name.
    """
    name = akw(0, args, str, None, invalid=("",), required=True)
    return str(join(CONFIG_DIR, name.strip() + ".cmake"))


def config_exists(*args):
    """Test if configuration of name exists.

    Args:
        name (str): Configuration name.
    """
    name = akw(0, args, str, None, invalid=("",), required=True)
    return bool(isfile(config_path(name)))


def config_new(*args):
    """Create configuration of name if it does not exist.

    Args:
        name (str): Configuration name.
    """
    name = akw(0, args, str, None, invalid=("",), required=True)
    if config_exists(name):
        return bool(False)

    # Copy template file to new location.
    try:
        copyfile(CONFIG_TEMPLATE_FILE, config_path(name))
    except IOError as e:
        return bool(False)

    return bool(True)


def config_delete(*args):
    """Delete existing configuration of name.

    Args:
        name (str): Configuration name.
    """
    name = akw(0, args, str, None, invalid=("",), required=True)
    if not config_exists(name):
        return bool(False)

    # Delete file.
    try:
        remove(config_path(name))
    except IOError as e:
        return bool(False)

    return bool(True)


def config_subparser(sb):
    """Populate subparser for config command.

    Arguments:
        sb (obj): argparse.ArgumentParser object.
    """
    # Mutually exclusive group of options for action.
    grp = sb.add_mutually_exclusive_group(required=True)

    # List existing configurations option.
    grp.add_argument("-l", "--list",
                     action="store_true",
                     help="list existing configurations")

    # Create new configuration option.
    grp.add_argument("-n", "--new",
                     nargs=1,
                     type=str,
                     help="create new configuration",
                     metavar="NAME")

    # Check configuration exists option.
    grp.add_argument("-e", "--exists",
                     nargs=1,
                     type=str,
                     help="check configuration exists",
                     metavar="NAME")

    # Delete existing configuration option.
    grp.add_argument("-d", "--delete",
                     nargs=1,
                     type=str,
                     help="delete existing configuration",
                     metavar="NAME")

    # Set callback function.
    sb.set_defaults(func=_subparser_func)


def _subparser_func(args):
    # Handle --list option.
    if args.list:
        msg = "config list\n{0}".format("\n".join(config_list()))
        return tuple((0, msg))

    # Handle --new option.
    elif args.new:
        name = args.new[0]
        if config_new(name):
            msg = "new config \"{0}\" ({1})".format(name, config_path(name))
            return tuple((0, msg))

        msg = "failed to create config \"{0}\"".format(name)
        return tuple((1, msg))

    # Handle --exists option.
    elif args.exists:
        name = args.exists[0]
        if config_exists(name):
            msg = "config exists \"{0}\" ({1})".format(name, config_path(name))
            return tuple((0, msg))

        msg = "config \"{0}\" does not exist".format(name)
        return tuple((1, msg))

    # Handle --delete option.
    elif args.delete:
        name = args.delete[0]
        if config_delete(name):
            msg = "config deleted \"{0}\"".format(name)
            return tuple((0, msg))

        msg = "failed to delete config \"{0}\"".format(name)
        return tuple((1, msg))

    # Should never reach here.
    return tuple((0, None))
